//#undef DEBUG
// ActualLineGraphView.cpp
// Copyright(C) 1999 by Kiyoshi Kobayashi. All rights reserved.
//
//
#include <Point.h>
#include <OS.h>
#include <Message.h>
#include <MenuItem.h>
#include <PopUpMenu.h>
#include <Application.h>
#include <Debug.h>

#include <stdio.h>
#include <string.h>

#include "MeterMessages.h"
#include "ActualLineGraphView.h"
#include "MeterBaseView.h"

static const char CLASS_NAME[] = "ActualLineGraphView";
static const char VIEW_NAME[] = "ActualLineGraphView";
static const char MAX_NAME[] = "ActualLineGraphView::defaultMax";

static const float GAP = 1.0f;
static const rgb_color LINE_COLOR = {20, 20, 20, 0};


// private class that keeps rate information.
class Rate {
public:
	Rate(float rate)
		: _rate(rate) {};
	virtual ~Rate() {};
	
	inline float GetRate() const { return _rate; };
	inline void SetRate(float newRate) { _rate = newRate; };
	
	inline BPoint GetPoint(float height, float x) {
		return BPoint(x, height - (height - 1.0) * _rate - 1.0);
	};

private:
	float _rate;
};

// Public Constructor.
//
ActualLineGraphView::ActualLineGraphView(BRect frame, const char *name,
					 float maximum)
	: BView(frame, name, B_FOLLOW_ALL, B_WILL_DRAW|B_FRAME_EVENTS),
	  max(maximum), defaultMax(maximum), menu(NULL) {}

// Public Constructor from the archive.
//
ActualLineGraphView::ActualLineGraphView(BMessage *archive)
	: BView(archive) {
	status_t ret;
	ret = archive->FindFloat(MAX_NAME, &defaultMax);
	ASSERT(ret == B_OK);
	max = defaultMax;
}

// Public Destructor.
//
ActualLineGraphView::~ActualLineGraphView() {}

// Public: To update the line graph with new value.
//
void
ActualLineGraphView::Update(float value) {
	float r = 0.0f;
	if (max != 0.0f) {
		if (max < value) {
			r = max / value;
			list.DoForEach(Regurate, (void *)&r);
			max = value;
		} else {
			r = value / max;
		}
	} else {
		if (value != 0.0f) {
			max = value;
			r = 1.0f;
		}
	}
	
	// prepare new data
	list.AddItem(new Rate(r), 0);
	PRINT(("lgvud: rate: %f, value: %f, max: %f\n", r, value, max));

	const int32 n = list.CountItems();
	if (n > num) {
		list.RemoveItems(num - 1, n - num);
	}
	
	// Update lines
	UpdateLines();
}

// Public: Override BView method.
//
void
ActualLineGraphView::AttachedToWindow() {
	if (Parent()) {
		num = (int)(Frame().Width() / GAP) + 4; // num of items in the list
		PRINT(("num: %d\n", num));

		SetViewColor(Parent()->ViewColor());

		// redraw all lines
		RedrawAllLines();
	} else
		TRESPASS();
	
	BView::AttachedToWindow();
}

// Public: Override BView method.
//
void
ActualLineGraphView::FrameResized(float width, float height) {
	// recalc num of items to be stored in the list.
	num = (int)(width / GAP) + 4;
	
	// redraw all lines
	RedrawAllLines();
	
	BView::FrameResized(width, height);
}

// Public: Override BView method.
//
void
ActualLineGraphView::Draw(BRect updateRect) {
	PRINT_OBJECT(updateRect);
	
	// redraw all lines
	RedrawAllLines();
	
	BView::Draw(updateRect);
}

// Public: Override BView method.
//
void
ActualLineGraphView::MouseDown(BPoint point) {
	// Popup the menu.
	BMenuItem *selected;
	ConvertToScreen(&point);
	selected = ActualLineGraphView::GetMenu()->Go(point, true, false, true);
}

// Public: Archive the object.
//
status_t
ActualLineGraphView::Archive(BMessage *archive, bool deep) const {
	TRACE();
	status_t ret;
	ret = BView::Archive(archive, deep);
	ASSERT(ret == B_OK);

	ret = archive->AddString("class", CLASS_NAME);
	ASSERT(ret == B_OK);

	if ( deep ) {
		ret = archive->AddFloat(MAX_NAME, defaultMax);
		ASSERT(ret == B_OK);
	}

	return B_OK;
}

// Public/Static: Instanciate the object.
//
BArchivable *
ActualLineGraphView::Instantiate(BMessage *archive) {
	if (validate_instantiation(archive, CLASS_NAME))
		return new ActualLineGraphView(archive); 
		
	return NULL; 
}

// Private: Copy constructor.
//
ActualLineGraphView::ActualLineGraphView(const ActualLineGraphView &)
	: BView(BRect(), NULL, B_FOLLOW_ALL, B_WILL_DRAW) {}

// Private: Redraw all lines.
//
void
ActualLineGraphView::RedrawAllLines() {
	// erase all
	FillRect(Bounds(), B_SOLID_LOW);
	
	const float height = Frame().Height();
	const float width = Frame().Width();
	
	const int items = list.CountItems();
	int i = 0;

	Rate *cur = (Rate *)list.ItemAt(i);
	Rate *pre = (Rate *)list.ItemAt(i + 1);

	// draw new lines
	BeginLineArray(items);
	while((pre != NULL) && (cur != NULL)) {
		BPoint curPoint = cur->GetPoint(height, width - GAP * (float)i);
		BPoint prePoint = pre->GetPoint(height, width - GAP * (float)(i + 1));
	
		AddLine(prePoint, curPoint, LINE_COLOR);
		i++;
		cur = pre;
		pre = (Rate *)list.ItemAt(i + 1);

	}

	EndLineArray();
}

// Private: Update lines (move bitmap and draw new line).
//
void
ActualLineGraphView::UpdateLines() {
	// shift existing lines
	BRect src(Bounds()), dst(Bounds());
	src.SetLeftTop(BPoint(src.left + GAP, src.top));
	dst.SetRightBottom(BPoint(dst.right - GAP, dst.bottom));
	CopyBits(src, dst);
	
	// clear remaining area
	src.SetLeftTop(BPoint(src.right, src.top));
	FillRect(src, B_SOLID_LOW);
		
	Rate *cur = (Rate *)list.ItemAt(0);
	Rate *pre = (Rate *)list.ItemAt(1);

	// draw new lines
	if ((pre != NULL) && (cur != NULL)) {
		const float width = Frame().Width();
		const float height = Frame().Height();
		BPoint curPoint = cur->GetPoint(height, width);
		BPoint prePoint = pre->GetPoint(height, width - GAP);
	
		StrokeLine(prePoint, curPoint);
	}
}

// Private: Construct PopUpMenu.
//
BPopUpMenu *
ActualLineGraphView::GetMenu() {
	if (menu != NULL)
		delete menu;
	
	menu = new BPopUpMenu("ActualLineGraphView Menu", false, false);
	BView *cur = this;
	BView *v;
	while ((v = cur->Parent()) != NULL) {
		if (is_kind_of(v, MeterBaseView)) {
			MeterBaseView *p = cast_as(v, MeterBaseView);
			ASSERT(p != NULL);
			p->AddMenuItems(menu);
		}
		cur = v;
	}
	return menu;
}

// Private: Regurate values in the list.
//
bool
ActualLineGraphView::Regurate(void *item, void *rate) {
	bool ret = false;
	if ((item != NULL) && (rate != NULL)) {
		Rate *f = (Rate *)item;
		float *r = (float *)rate;
		f->SetRate(f->GetRate() * (*r));
	} else
		ret = true;
	
	return ret;
}

